package com.ningdatech.pmapi.meeting.manage;

import cn.hutool.core.bean.BeanUtil;
import cn.hutool.core.collection.CollUtil;
import cn.hutool.core.util.ObjectUtil;
import cn.hutool.core.util.StrUtil;
import cn.hutool.crypto.SecureUtil;
import com.alibaba.fastjson.JSON;
import com.baomidou.mybatisplus.core.conditions.query.LambdaQueryWrapper;
import com.baomidou.mybatisplus.core.conditions.update.LambdaUpdateWrapper;
import com.baomidou.mybatisplus.core.toolkit.Wrappers;
import com.baomidou.mybatisplus.extension.plugins.pagination.Page;
import com.ningdatech.basic.exception.BizException;
import com.ningdatech.basic.model.IdVo;
import com.ningdatech.basic.model.PageVo;
import com.ningdatech.basic.util.CollUtils;
import com.ningdatech.basic.util.ValidUtil;
import com.ningdatech.cache.lock.DistributedLock;
import com.ningdatech.file.service.FileService;
import com.ningdatech.pmapi.common.helper.RegionCacheHelper;
import com.ningdatech.pmapi.expert.entity.ExpertUserFullInfo;
import com.ningdatech.pmapi.expert.helper.PermissionCheckHelper;
import com.ningdatech.pmapi.expert.service.IExpertUserFullInfoService;
import com.ningdatech.pmapi.meeting.builder.ExpertInviteBuilder;
import com.ningdatech.pmapi.meeting.entity.domain.*;
import com.ningdatech.pmapi.meeting.entity.dto.*;
import com.ningdatech.pmapi.meeting.entity.enumeration.ExpertAttendStatus;
import com.ningdatech.pmapi.meeting.entity.enumeration.ExpertInviteTypeEnum;
import com.ningdatech.pmapi.meeting.entity.enumeration.MeetingStatus.Manager;
import com.ningdatech.pmapi.meeting.entity.req.*;
import com.ningdatech.pmapi.meeting.entity.vo.*;
import com.ningdatech.pmapi.meeting.entity.vo.ExpertInviteDetailVO.ExpertAttendListItemVO;
import com.ningdatech.pmapi.meeting.entity.vo.ExpertInviteDetailVO.RandomInviteListItemVO;
import com.ningdatech.pmapi.meeting.helper.ExpertInviteHelper;
import com.ningdatech.pmapi.meeting.helper.MeetingManageHelper;
import com.ningdatech.pmapi.meeting.helper.YxtCallOrSmsHelper;
import com.ningdatech.pmapi.meeting.service.*;
import com.ningdatech.pmapi.meeting.task.ExpertInviteTask;
import com.ningdatech.pmapi.meta.helper.DictionaryCache;
import com.ningdatech.pmapi.meta.helper.TagCache;
import com.ningdatech.pmapi.sys.model.dto.RegionDTO;
import com.ningdatech.pmapi.user.security.auth.model.UserInfoDetails;
import com.ningdatech.pmapi.user.service.IUserInfoService;
import com.ningdatech.pmapi.user.util.LoginUserUtil;
import lombok.RequiredArgsConstructor;
import org.springframework.stereotype.Component;
import org.springframework.transaction.annotation.Transactional;
import org.springframework.util.Assert;

import java.time.LocalDateTime;
import java.util.*;
import java.util.function.Function;
import java.util.stream.Collectors;

import static com.ningdatech.pmapi.meeting.helper.ExpertInviteHelper.getExpertInviteRule;

/**
 * <p>
 * MeetingManage
 * </p>
 *
 * @author WendyYang
 * @since 13:35 2022/7/26
 */
@Component
@RequiredArgsConstructor
public class MeetingManage {

    private final RegionCacheHelper regionCache;
    private final IMeetingService meetingService;
    private final IExpertInviteAvoidRuleService inviteAvoidRuleService;
    private final IExpertInviteRuleService inviteRuleService;
    private final IExpertUserFullInfoService expertUserFullInfoService;
    private final FileService fileService;
    private final TagCache tagCache;
    private final DictionaryCache dictionaryCache;
    private final IMeetingExpertService meetingExpertService;
    private final ExpertInviteManage expertInviteManage;
    private final ExpertInviteTask expertInviteTask;
    private final MeetingManageHelper meetingManageHelper;
    private final YxtCallOrSmsHelper yxtCallOrSmsHelper;
    private final DistributedLock distributedLock;
    private final PermissionCheckHelper permissionCheckHelper;
    private final IUserInfoService userInfoService;
    private final IMeetingInnerProjectService meetingInnerProjectService;
    private final IMeetingOuterProjectService meetingOuterProjectService;

    private final ExpertInviteHelper expertInviteHelper;
    private static final String INVITED_RULE_CREATE = "INVITED_RULE_CREATE:";

    private static final String MEETING_CREATE_KEY = "MEETING_CREATE:";

    /**
     * 锁定时间10 * 60秒
     */
    public static final int RETRY_TIMES = 3;

    @Transactional(rollbackFor = Exception.class)
    public IdVo<Long> meetingCreateAndInviteExpert(MeetingCreateReq req) {
        MeetingBasicDTO meetingBasic = req.getMeetingBasicInfo();
        meetingManageHelper.checkReviewProject(meetingBasic);
        String md5ByParam = SecureUtil.md5(meetingBasic.toString());
        String key = MEETING_CREATE_KEY + md5ByParam;
        if (!distributedLock.lock(key, RETRY_TIMES)) {
            throw BizException.wrap("会议正在创建中");
        }
        try {
            // 保存会议基本信息
            Meeting meeting = BeanUtil.copyProperties(meetingBasic, Meeting.class);
            meeting.setStatus(Manager.UNCOMPLETED.getCode());
            UserInfoDetails userDetail = LoginUserUtil.loginUserDetail();
            meeting.setHoldOrg(userDetail.getOrganizationName());
            meeting.setHoldOrgCode(userDetail.getOrganizationCode());
            meeting.setRegionCode(userDetail.getRegionCode());
            meeting.setCreator(userDetail.getUsername());
            meeting.setInviteStopped(true);
            meetingService.save(meeting);
            if (meetingBasic.getIsInnerProject()) {
                List<MeetingInnerProject> projects = meetingBasic.getProjectIdList().stream().map(w -> {
                    MeetingInnerProject project = new MeetingInnerProject();
                    project.setMeetingId(meeting.getId());
                    project.setProjectId(w);
                    return project;
                }).collect(Collectors.toList());
                meetingInnerProjectService.saveBatch(projects);
            } else {
                List<MeetingOuterProject> projects = meetingBasic.getProjects().stream().map(w -> {
                    MeetingOuterProject project = BeanUtil.copyProperties(w, MeetingOuterProject.class);
                    project.setMeetingId(meeting.getId());
                    return project;
                }).collect(Collectors.toList());
                meetingOuterProjectService.saveBatch(projects);
            }
            // 抽取专家
            req.getExpertInviteRule().setMeetingId(meeting.getId());
            expertInviteByCreate(req.getExpertInviteRule());
            return IdVo.of(meeting.getId());
        } finally {
            distributedLock.releaseLock(key);
        }
    }

    @Transactional(rollbackFor = Exception.class)
    public void expertInviteByCreate(ExpertInviteReq req) {
        String key = INVITED_RULE_CREATE + req.getMeetingId();
        if (!distributedLock.lock(key, RETRY_TIMES)) {
            throw BizException.wrap("不可重复进行专家抽取");
        }
        try {
            Meeting meeting = meetingService.getById(req.getMeetingId());
            if (ExpertInviteTypeEnum.RANDOM.eq(req.getInviteType())) {
                List<RandomInviteRuleDTO> randomRules = req.getRandomRules();
                Assert.notEmpty(randomRules, "随机抽取规则不能为空");
                AvoidInfoDTO avoidInfo = req.getAvoidRule();
                Assert.notNull(avoidInfo, "回避信息不能为空");
                // 随机抽取的话则需进行抽取数量校验
                ExpertCountOnChangeVO countOnChange = expertCountOnChange(req);
                if (!countOnChange.getStatus()) {
                    throw BizException.wrap(countOnChange.getMessage());
                }
                for (int i = 0; i < randomRules.size(); i++) {
                    Integer checkCount = countOnChange.getCountList().get(i);
                    Integer inviteCount = randomRules.get(i).getCount();
                    Assert.isTrue(checkCount >= inviteCount, "可供抽取的专家数量不足");
                }
                expertInviteManage.expertRandomInviteByMeetingCreate(meeting, randomRules, avoidInfo);
                expertInviteTask.addInviteExpertTaskByMeetingCreate(meeting.getId(), 5);
                LambdaUpdateWrapper<Meeting> update = Wrappers.lambdaUpdate(Meeting.class);
                update.set(Meeting::getInviteStopped, false);
                update.eq(Meeting::getId, meeting.getId());
                meetingService.update(update);
                // 回避规则
                if (avoidInfo != null) {
                    ExpertInviteAvoidRule avoidRule = new ExpertInviteAvoidRule();
                    avoidRule.setMeetingId(meeting.getId());
                    // 未传值时设置为0 表示不限制周参与次数
                    avoidRule.setWeekInviteCount(ObjectUtil.defaultIfNull(avoidRule.getWeekInviteCount(), 0));
                    avoidRule.setAvoidOrgIds(CollUtils.joinByComma(avoidInfo.getAvoidOrgIdList()));
                    avoidRule.setAvoidUnitIds(CollUtils.joinByComma(avoidInfo.getAvoidUnitIdList()));
                    avoidRule.setExpertIds(CollUtils.joinByComma(avoidInfo.getExpertIds()));
                    inviteAvoidRuleService.save(avoidRule);
                }
            } else {
                // 指定邀请
                AppointInviteRuleDTO appointRule = req.getAppointRule();
                Assert.notNull(appointRule, "指定邀请规则不能为空");
                Map<Long, ExpertUserFullInfo> usersMap = expertInviteHelper.checkAppointExpert(appointRule);
                // 处理指定抽取规则
                ExpertInviteRule appoint = getExpertInviteRule(appointRule, meeting.getId());
                inviteRuleService.save(appoint);
                Long ruleId = appoint.getId();
                List<MeetingExpert> experts = appointRule.getExpertIds().stream().map(w -> {
                    ExpertUserFullInfo info = usersMap.get(w);
                    MeetingExpert expert = ExpertInviteBuilder.getExpertByAppoint(meeting.getId(), info, ruleId);
                    expert.setStatus(ExpertAttendStatus.NOTICING.getCode());
                    return expert;
                }).collect(Collectors.toList());
                meetingExpertService.saveBatch(experts);
            }
        } finally {
            distributedLock.releaseLock(key);
        }
    }

    /**
     * 抽取数量校验
     *
     * @param req 抽取参数
     * @return {@link ExpertCountOnChangeVO}
     * @author WendyYang
     **/
    public ExpertCountOnChangeVO expertCountOnChange(ExpertInviteReq req) {
        // 未传递指定邀请专家直接忽略
        expertInviteHelper.checkAppointExpert(req.getAppointRule());
        Meeting meeting = meetingService.getById(req.getMeetingId());
        ExpertCountOnChangeVO resultCount = new ExpertCountOnChangeVO();
        resultCount.setStatus(true);
        resultCount.setCountList(new ArrayList<>());
        // 参数校验
        String errMsg = ValidUtil.validFast(req, AbstractInviteRule.CountCheck.class);
        if (errMsg != null) {
            resultCount.setStatus(false);
            resultCount.setMessage(errMsg);
            return resultCount;
        }
        List<Long> expertIdsChoose = new ArrayList<>();
        for (RandomInviteRuleDTO randomRule : req.getRandomRules()) {
            ExpertChooseDTO chooseExpert = expertInviteManage.expertInviteByRandomRule(req.getAvoidRule(),
                    randomRule, expertIdsChoose, meeting.getStartTime(), meeting.getEndTime());
            resultCount.addCountList(chooseExpert.getTotal());
        }
        return resultCount;
    }

    /**
     * 专家会议列表
     *
     * @param req 查询参数
     * @return 会议列表
     * @author WendyYang
     **/
    public PageVo<MeetingByManagerVO> meetingListByExpert(MeetingListReq req) {
        Long expertId = req.getExpertId() != null ? req.getExpertId() : LoginUserUtil.getUserId();
        List<MeetingAndAttendStatusDTO> expertDtoList = meetingExpertService.listByExpertIdAndStatus(expertId, req.getStatus(), null);
        if (expertDtoList.isEmpty()) {
            return PageVo.empty();
        }
        Map<Long, MeetingAndAttendStatusDTO> mapByMeetingId = new HashMap<>(16);
        expertDtoList.forEach(w -> mapByMeetingId.put(w.getMeetingId(), w));
        LambdaQueryWrapper<Meeting> query = new LambdaQueryWrapper<Meeting>()
                .orderByDesc(Meeting::getCreateOn)
                .in(Meeting::getId, mapByMeetingId.keySet())
                .ne(Meeting::getStatus, Manager.CANCELED.getCode());
        if (req.getExpertId() == null) {
            meetingManageHelper.buildMeetingQuery(query, req);
        }
        Page<Meeting> page = meetingService.page(req.page(), query);
        if (page.getTotal() == 0) {
            return PageVo.empty();
        }
        PageVo<MeetingByManagerVO> result = new PageVo<>(new ArrayList<>(), page.getTotal());
        page.getRecords().forEach(meeting -> {
            MeetingByManagerVO item = meetingManageHelper.buildByMeeting(meeting);
            MeetingAndAttendStatusDTO info = mapByMeetingId.get(meeting.getId());
            item.setAttendStatus(meetingManageHelper.getExpertAttendStatus(info));
            result.getRecords().add(item);
        });
        return result;
    }

    /**
     * 管理员会议列表
     *
     * @param po 查询参数
     * @return 会议列表
     * @author WendyYang
     */
    public PageVo<MeetingByManagerVO> meetingListByManager(MeetingListReq po) {
        LambdaQueryWrapper<Meeting> query = new LambdaQueryWrapper<Meeting>()
                .orderByDesc(Meeting::getCreateOn);
        // 补充逻辑 如果拥有超级管理员权限可以看到所有事务
        if (!permissionCheckHelper.isSuperAdmin()) {
            query.eq(Meeting::getCreateBy, LoginUserUtil.getUserId());
        }
        query.eq(po.getStatus() != null, Meeting::getStatus, po.getStatus());
        meetingManageHelper.buildMeetingQuery(query, po);
        Page<Meeting> page = meetingService.page(po.page(), query);
        if (page.getTotal() == 0) {
            return PageVo.empty();
        }
        PageVo<MeetingByManagerVO> result = new PageVo<>(new ArrayList<>(), page.getTotal());
        Map<Long, CountConfirmByMeetingIdDTO> countConfirmMap = new HashMap<>(16);
        List<Long> meetingIds = CollUtils.fieldList(page.getRecords(), Meeting::getId);
        countConfirmMap.putAll(meetingExpertService.countConfirmedByMeetingIds(meetingIds));
        page.getRecords().forEach(meeting -> {
            MeetingByManagerVO item = meetingManageHelper.buildByMeeting(meeting);
            CountConfirmByMeetingIdDTO confirm = countConfirmMap.get(meeting.getId());
            if (confirm == null) {
                item.setInviteCount(0);
                item.setConfirmCount(0);
            } else {
                item.setInviteCount(confirm.getTotal());
                item.setConfirmCount(confirm.getConfirmed());
            }
            result.getRecords().add(item);
        });
        return result;
    }

    public MeetingDetailBasicVO getMeetingBasicInfo(Long meetingId) {
        Meeting meeting = meetingService.getById(meetingId);
        Assert.notNull(meeting, "会议不存在");
        return MeetingDetailBasicVO.builder()
                .meetingId(meeting.getId())
                .meetingName(meeting.getName())
                .meetingType(meeting.getType())
                .meetingAddress(meeting.getMeetingAddress())
                .typeName(dictionaryCache.getByCode(meeting.getType()).getName())
                .startTime(meeting.getStartTime())
                .judgesAttendanceTime(meeting.getJudgesAttendanceTime())
                .meetingUsageTime(meeting.getMeetingUsageTime())
                .contact(meeting.getContact())
                .connecter(meeting.getConnecter())
                .status(meeting.getStatus())
                .holdOrg(meeting.getHoldOrg())
                .createOn(meeting.getCreateOn())
                .creator(meeting.getCreator())
                .createBy(meeting.getCreator())
                .build();
    }

    public ExpertInviteDetailVO inviteDetail(Long meetingId) {
        ExpertInviteDetailVO result = new ExpertInviteDetailVO();
        Meeting meeting = meetingService.getById(meetingId);
        if (Objects.isNull(meeting)) {
            throw new BizException("该会议信息不存在");
        }
        result.setInvitedStopped(meeting.getInviteStopped());
        List<MeetingExpert> experts = meetingExpertService.listByMeetingId(meetingId);
        if (experts.isEmpty()) {
            return result;
        }
        List<MeetingExpert> randomList = new ArrayList<>();
        List<MeetingExpert> appointList = new ArrayList<>();
        List<MeetingExpert> attendList = new ArrayList<>();
        List<Long> expertIds = new ArrayList<>();
        experts.forEach(w -> {
            boolean randomInvite = w.getInviteType().equals(ExpertInviteTypeEnum.RANDOM.getCode());
            if (randomInvite) {
                randomList.add(w);
            } else {
                appointList.add(w);
            }
            if (w.getStatus().equals(ExpertAttendStatus.AGREED.getCode())) {
                attendList.add(w);
                if (randomInvite) {
                    result.setRandomAttend(result.getRandomAttend() + 1);
                } else {
                    result.setAppointAttend(result.getAppointAttend() + 1);
                }
            }
            expertIds.add(w.getExpertId());
        });
        result.setAttendTotal(attendList.size());
        Map<Long, ExpertBasicInfoVO> expertBasicInfoVoMap = meetingManageHelper.getExpertBasicInfo(expertIds);
        Function<MeetingExpert, RandomInviteListItemVO> mapping = sme -> {
            ExpertBasicInfoVO basicInfoVo = expertBasicInfoVoMap.get(sme.getExpertId());
            RandomInviteListItemVO item = BeanUtil.copyProperties(basicInfoVo, RandomInviteListItemVO.class);
            item.setStatus(sme.getStatus());
            item.setMeetingId(sme.getMeetingId());
            item.setExpertMeetingId(sme.getId());
            ExpertAttendStatus status = ExpertAttendStatus.getByCode(sme.getStatus());
            if (status.equals(ExpertAttendStatus.NOTICING)) {
                item.setNoticeStatus(status.getDesc());
                item.setConfirmResult(StrUtil.EMPTY);
            } else {
                item.setNoticeStatus("已通知");
                item.setConfirmResult(status.getDesc());
            }
            return item;
        };
        // 随机邀请列表
        randomList.forEach(w -> result.addRandomInviteList(mapping.apply(w)));
        // 指定抽取列表
        appointList.forEach(w -> result.addAppointInviteList(mapping.apply(w)));
        // 确定参加列表
        attendList.forEach(w -> {
            ExpertBasicInfoVO expertBasicInfoVo = expertBasicInfoVoMap.get(w.getExpertId());
            ExpertAttendListItemVO item = BeanUtil.copyProperties(expertBasicInfoVo, ExpertAttendListItemVO.class);
            item.setInviteType(ExpertInviteTypeEnum.getByCode(w.getInviteType()).getName());
            result.addAttendList(item);
        });
        return result;
    }

    private boolean meetingInfoChange(Meeting old, Meeting current) {
        if (!old.getStartTime().equals(current.getStartTime())) {
            return Boolean.TRUE;
        }
        if (!old.getRegionCode().equals(current.getRegionCode())) {
            return Boolean.TRUE;
        }
        return Boolean.FALSE;
    }

    /**
     * 会议基本信息修改
     *
     * @param po 修改参数
     **/
    public void meetingBasicInfoModify(MeetingBasicInfoModifyReq po) {
        Meeting meeting = new Meeting();
        BeanUtil.copyProperties(po, meeting);
        LocalDateTime now = LocalDateTime.now();
        Meeting old = meetingService.getById(po.getId());
        if (Manager.UNCOMPLETED.eq(old.getStatus()) && meetingInfoChange(old, meeting)) {
            List<MeetingExpert> meList = meetingExpertService.listExpertByAgreeAttend(Collections.singletonList(po.getId()));
            if (!meList.isEmpty() && old.getStartTime().isAfter(now)) {
                // TODO
                /*String meetingType = dictionaryCache.getByCode(old.getType()).getName();
                List<SendSmsContext> contexts = YxtSmsContextBuilder.smsToExpertByMeetingChange(old, meeting, meList, meetingType);
                yxtCallOrSmsHelper.sendSms(contexts);*/
            }
        }
        meetingService.updateById(meeting);
    }

    /**
     * 抽取规则详情
     *
     * @param meetingId 会议ID
     * @return {@link InviteRuleDetailVO}
     * @author WendyYang
     **/
    public InviteRuleDetailVO inviteRuleDetail(Long meetingId) {
        InviteRuleDetailVO result = new InviteRuleDetailVO();
        List<ExpertInviteRule> inviteRules = inviteRuleService.listByMeetingId(meetingId);
        result.setRandomRules(new ArrayList<>());
        Map<ExpertInviteTypeEnum, List<ExpertInviteRule>> groupByType = CollUtils.group(inviteRules, w -> ExpertInviteTypeEnum.getByCode(w.getInviteType()));
        List<ExpertInviteRule> randoms = groupByType.get(ExpertInviteTypeEnum.RANDOM);
        if (CollUtil.isNotEmpty(randoms)) {
            randoms.forEach(random -> {
                RandomInviteRuleVO randomRule = JSON.parseObject(random.getInviteRule(), RandomInviteRuleVO.class);
                if (randomRule.getExpertTags() != null) {
                    randomRule.getExpertTags().forEach(w -> {
                        List<String> tagNames = CollUtils.convert(w.getTagCodes(), tagCode -> tagCache.getByTagCode(tagCode).getTagName());
                        w.setTagCodes(tagNames);
                    });
                }
                if (randomRule.getExpertDicts() != null) {
                    randomRule.getExpertDicts().forEach(w -> {
                        List<String> dictNames = CollUtils.convert(w.getDictCodes(), dictCode -> dictionaryCache.getByCode(dictCode).getName());
                        w.setDictCodes(dictNames);
                    });
                }
                if (StrUtil.isNotEmpty(randomRule.getIntentionRegionCode())) {
                    List<RegionDTO> intentionRegions = regionCache.listParents(randomRule.getIntentionRegionCode(), randomRule.getIntentionRegionLevel());
                    randomRule.setIntentionRegions(intentionRegions);
                }
                if (StrUtil.isNotEmpty(randomRule.getExpertRegionCode())) {
                    List<RegionDTO> expertRegions = regionCache.listParents(randomRule.getExpertRegionCode(), randomRule.getExpertRegionLevel());
                    randomRule.setExpertRegions(expertRegions);
                }
                result.getRandomRules().add(randomRule);
            });
            AvoidInfoDTO avoidInfo = inviteAvoidRuleService.getAvoidInfoDto(meetingId);
            AvoidInfoVO vo = new AvoidInfoVO();
            vo.setAvoidOrgIds(avoidInfo.getAvoidOrgIdList());
            vo.setAvoidUnitIds(avoidInfo.getAvoidUnitIdList());
            if (CollUtil.isNotEmpty(avoidInfo.getExpertIds())) {
                vo.setExperts(new ArrayList<>(meetingManageHelper.getExpertBasicInfo(avoidInfo.getExpertIds()).values()));
            }
            result.setAvoidInfo(vo);
        } else {
            List<ExpertInviteRule> appoints = groupByType.get(ExpertInviteTypeEnum.APPOINT);
            ExpertInviteRule appoint = appoints.get(0);
            AppointInviteRuleDTO appointRule = JSON.parseObject(appoint.getInviteRule(), AppointInviteRuleDTO.class);
            AppointRuleVO vo = new AppointRuleVO();
            vo.setInviteDesc(appointRule.getInviteDesc());
            vo.setExperts(new ArrayList<>(meetingManageHelper.getExpertBasicInfo(appointRule.getExpertIds()).values()));
            result.setAppointRule(vo);
        }
        return result;
    }

    public void expertRemove(ExpertRemoveReq po) {
        LambdaUpdateWrapper<MeetingExpert> update = Wrappers.lambdaUpdate(MeetingExpert.class)
                .eq(MeetingExpert::getId, po.getExpertMeetingId())
                .set(MeetingExpert::getStatus, ExpertAttendStatus.CANCELED.getCode());
        meetingExpertService.update(update);
    }

    @Transactional(rollbackFor = Exception.class)
    public void expertReplace(ExpertRemoveReq po) {
        MeetingExpert meetingExpert = meetingExpertService.getById(po.getExpertMeetingId());
        ExpertUserFullInfo expertFullInfo;
        Long ruleId = 0L;
        Meeting meeting = meetingService.getById(po.getMeetingId());
        if (po.getExpertId() != null) {
            // 指定邀请替换
            List<ExpertUserFullInfo> userInfos = meetingManageHelper.appointExpertCheck(po.getMeetingId(), Collections.singletonList(po.getExpertId()));
            expertFullInfo = userInfos.get(0);
        } else {
            List<ExpertInviteRule> inviteRules = inviteRuleService.listByMeetingId(po.getMeetingId());
            // 邀请规则
            RandomInviteRuleDTO randomInviteRuleDto = null;
            for (ExpertInviteRule rule : inviteRules) {
                if (rule.getInviteType().equals(ExpertInviteTypeEnum.RANDOM.getCode()) &&
                        rule.getId().equals(meetingExpert.getRuleId())) {
                    randomInviteRuleDto = JSON.parseObject(rule.getInviteRule(), RandomInviteRuleDTO.class);
                    ruleId = rule.getId();
                    break;
                }
            }
            // 回避规则
            AvoidInfoDTO avoidInfoDto = meetingManageHelper.getAvoidInfoDto(po.getMeetingId());
            // 添加回避该替换的专家
            List<Long> expertIds = avoidInfoDto.getExpertIds();
            if (CollUtil.isEmpty(expertIds)) {
                expertIds = new ArrayList<>();
            }
            expertIds.add(meetingExpert.getExpertId());
            avoidInfoDto.setExpertIds(expertIds);
            meetingManageHelper.saveAvoidInfo(po.getMeetingId(), avoidInfoDto.getExpertIds());

            List<MeetingExpert> meetingExperts = meetingExpertService.listByMeetingId(po.getMeetingId());
            ExpertChooseDTO expertChooseDto = expertInviteManage.expertReplaceByRandomRule(avoidInfoDto, randomInviteRuleDto,
                    meetingExperts, 1, meeting.getStartTime(), meeting.getEndTime(), meetingExpert.getExpertId());
            Assert.isTrue(expertChooseDto.getTotal() > 0, "暂无专家可供替换");
            expertFullInfo = expertChooseDto.getExperts().get(0);
            expertInviteTask.addInviteExpertTask(po.getMeetingId());
        }
        LambdaUpdateWrapper<MeetingExpert> update = Wrappers.lambdaUpdate(MeetingExpert.class)
                .eq(MeetingExpert::getId, po.getExpertMeetingId())
                .set(MeetingExpert::getUpdateOn, LocalDateTime.now())
                .set(MeetingExpert::getPreStatus, meetingExpert.getStatus())
                .set(MeetingExpert::getStatus, ExpertAttendStatus.REPLACED.getCode());
        meetingExpertService.update(update);
        MeetingExpert me;
        if (po.getExpertId() == null) {
            me = ExpertInviteBuilder.getExpertByRandom(po.getMeetingId(), expertFullInfo, ruleId);
        } else {
            me = ExpertInviteBuilder.getExpertByAppoint(po.getMeetingId(), expertFullInfo, ruleId);
        }
        me.setStatus(ExpertAttendStatus.NOTICING.getCode());
        me.setPreId(po.getExpertMeetingId());
        yxtCallOrSmsHelper.callByMeetingExperts(meeting, Collections.singletonList(me));
        meetingExpertService.save(me);
        // 发送专家替换短信 TODO
        // String meetingType = dictionaryCache.getByCode(meeting.getType()).getName();
        // SendSmsContext context = YxtSmsContextBuilder.smsToExpertByReplace(meeting, meetingExpert, meetingType);
        // yxtCallOrSmsHelper.sendSms(context);
    }

    public void batchAppointExperts(BatchAppointExpertsReq po) {
        List<ExpertUserFullInfo> userInfos = meetingManageHelper.appointExpertCheck(po.getMeetingId(), po.getExpertIds());
        List<MeetingExpert> expertList = CollUtils.convert(userInfos, w -> {
            MeetingExpert me = ExpertInviteBuilder.getExpertByAppoint(po.getMeetingId(), w, 0L);
            me.setStatus(ExpertAttendStatus.NOTICING.getCode());
            return me;
        });
        meetingExpertService.saveBatch(expertList);
    }

    public void sendMeetingNotice(Long meetingId) {
        String key = "SEND_MEETING_NOTICE:" + meetingId;
        if (!distributedLock.lock(key, RETRY_TIMES)) {
            throw BizException.wrap("正在下发会议通知");
        }
        try {
            Meeting meeting = meetingService.getById(meetingId);
            Assert.isTrue(Manager.UNCOMPLETED.eq(meeting.getStatus()), "非未完成会议无法发送会议通知");
            Assert.isTrue(meeting.getInviteStopped(), "随机邀请未结束");
            int noticeCount = meetingExpertService.countExpertByStatusAndMeetingId(ExpertAttendStatus.NOTICING, meetingId, null);
            Assert.isTrue(noticeCount == 0, "存在未确认完成的专家，暂无法下发会议通知");
            LambdaUpdateWrapper<Meeting> update = Wrappers.lambdaUpdate(Meeting.class)
                    .eq(Meeting::getId, meetingId);
            meetingService.update(update);
            // 发送会议通知
            List<MeetingExpert> experts = meetingExpertService.listExpertByAgreeAttend(Collections.singletonList(meetingId));
            if (!experts.isEmpty()) {
                // TODO
                // String meetingType = dictionaryCache.getByCode(meeting.getType()).getName();
                // List<SendSmsContext> contexts = YxtSmsContextBuilder.smsToExpertBySendNotice(meeting, experts, meetingType);
                // yxtCallOrSmsHelper.sendSms(contexts);
            }
        } finally {
            distributedLock.releaseLock(key);
        }
    }

    public void stopRandomInvite(Long meetingId) {
        expertInviteTask.cancelByMeetingId(meetingId);
    }

    @Transactional(rollbackFor = Exception.class)
    public void cancelMeeting(MeetingCancelReq po) {
        String key = "CANCEL_MEETING:" + po.getMeetingId();
        if (!distributedLock.lock(key, RETRY_TIMES)) {
            throw BizException.wrap("正在取消会议，请刷新后重试");
        }
        try {
            Meeting meeting = meetingService.getById(po.getMeetingId());
            Assert.isTrue(!Manager.CANCELED.eq(meeting.getStatus()), "会议已取消");
            Assert.isTrue(meeting.getStartTime().isAfter(LocalDateTime.now()), "会议已开始，暂时无法取消");
            LambdaUpdateWrapper<Meeting> update = Wrappers.lambdaUpdate(Meeting.class)
                    .set(Meeting::getStatus, Manager.CANCELED.getCode())
                    .set(Meeting::getUpdateBy, LoginUserUtil.getUserId())
                    .set(Meeting::getUpdateOn, LocalDateTime.now())
                    .eq(Meeting::getId, po.getMeetingId());
            meetingService.update(update);
            expertInviteTask.cancelByMeetingId(po.getMeetingId());
            // 发送通知给专家
            List<MeetingExpert> experts = meetingExpertService.listExpertByAgreeAttend(Collections.singletonList(po.getMeetingId()));
            if (!experts.isEmpty()) {
                // TODO
                // meeting.setCancelRemark(po.getCancelRemark());
                // String meetingType = dictionaryCache.getByCode(meeting.getType()).getName();
                // List<SendSmsContext> contexts = YxtSmsContextBuilder.smsToExpertByCancelMeeting(meeting, experts, meetingType);
                // yxtCallOrSmsHelper.sendSms(contexts);
            }
        } finally {
            distributedLock.releaseLock(key);
        }
    }

    public ExpertInvitationDetailVO expertInvitationDetail(Long meetingId, Long expertId) {
        Long userId = expertId == null ? LoginUserUtil.getUserId() : expertId;
        MeetingExpert me = meetingExpertService.getByMeetingIdAndExpertId(meetingId, userId);
        Assert.notNull(me, "未被邀请参加");
        Assert.isTrue(ExpertAttendStatus.AGREED.eq(me.getStatus()), "未确认参加");
        ExpertUserFullInfo expertInfo = expertUserFullInfoService.getByUserId(userId);
        Meeting meeting = meetingService.getById(meetingId);
        return ExpertInvitationDetailVO.builder()
                .expertName(expertInfo.getExpertName())
                .holdOrg(meeting.getHoldOrg())
                .meetingName(meeting.getName())
                .startTime(meeting.getStartTime())
                .endTime(meeting.getEndTime())
                .inviteTime(me.getCreateOn())
                .connecter(meeting.getConnecter())
                .contact(meeting.getContact())
                .build();
    }

    public void confirmAttendByManager(ExpertRemoveReq po) {
        MeetingExpert meetingExpert = meetingExpertService.getById(po.getExpertMeetingId());
        if (meetingExpert.getStatus().equals(ExpertAttendStatus.NOTICING.getCode())) {
            LambdaUpdateWrapper<MeetingExpert> update = Wrappers.lambdaUpdate(MeetingExpert.class)
                    .set(MeetingExpert::getStatus, ExpertAttendStatus.AGREED.getCode())
                    .set(MeetingExpert::getUpdateOn, LocalDateTime.now())
                    .set(MeetingExpert::getUpdateBy, LoginUserUtil.getUserId())
                    .eq(MeetingExpert::getId, po.getExpertMeetingId());
            meetingExpertService.update(update);
        } else {
            ExpertAttendStatus status = ExpertAttendStatus.getByCode(meetingExpert.getStatus());
            throw BizException.wrap("该专家" + status.getDesc());
        }
    }

}
